# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import pkg_resources
import logging

from knockoff.exceptions import ResourceNotFoundError
from knockoff.exceptions import NoEntryPointGroupError

logger = logging.getLogger(__name__)


# noinspection PyAttributeOutsideInit
class ResourceLocatorMixin(object):

    entry_point_group = None

    @property
    def registered_resources(self):
        try:
            return isinstance(self.__resources, dict)
        except AttributeError:
            return False

    def lazy_register_resources(self):
        if not self.registered_resources:
            self.register_resources()

    def register_resources(self):
        if self.entry_point_group is None:
            raise NoEntryPointGroupError(f"No entry_point declared for {self}")
        if self.registered_resources:
            logger.warning("Resources already registered. "
                           "Registering again will override "
                           "existing resources")
        self.__resources = {}
        for entry_point in (pkg_resources
                            .iter_entry_points(self
                                               .entry_point_group)):
            self.__resources[entry_point.name] = entry_point.load()

    def get_resource(self, name):
        self.lazy_register_resources()
        try:
            return self.__resources[name]
        except KeyError:
            raise ResourceNotFoundError(f"Resource not found: {name}"
                                        f" (entry_point_group="
                                        f"{self.entry_point_group})")

    def get_resource_names(self):
        self.lazy_register_resources()
        return self.__resources.keys()


def noop(*args, **kwargs):
    pass



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import six
import logging
from abc import ABCMeta, abstractmethod
from sqlalchemy import MetaData, Table, inspect
from sqlalchemy.types import JSON
from sqlalchemy.exc import NoSuchTableError

from faker import Faker
from numpy import random

from knockoff.orm import get_engine, get_child_tables
from .constraints import KnockoffUniqueConstraint
from .dag import DagService, Node

logger = logging.getLogger(__name__)


class KnockoffDatabaseService(six.with_metaclass(ABCMeta, object)):
    @abstractmethod
    def reflect_table(self, name):
        return

    @abstractmethod
    def insert(self, name, df, dtype=None):
        return

    @abstractmethod
    def reflect_unique_constraints(self, name):
        return

    @abstractmethod
    def has_table(self, name):
        return


class DefaultDatabaseService(KnockoffDatabaseService):
    def __init__(self, engine=None, **kwargs):
        self.engine = engine or get_engine()
        self.kwargs = {
            'if_exists': 'append',
            'index': False,
            'method': 'multi'
        }
        self.kwargs.update(kwargs)

    def has_table(self, name):
        with self.engine.connect() as conn:
            return conn.dialect.has_table(conn, name)

    def _resolve_table_name(self, name):
        with self.engine.connect() as conn:
            inspector = inspect(conn)
            try:
                inspector.get_table_oid(name)
                return name
            except NoSuchTableError:
                # if this is a postgres partitioned table, we'll try to load
                # with a child table
                logger.warning("Got NoSuchTableError for {}. "
                               "Will attempt to get child partitions "
                               "in case this is a postgres partitioned "
                               "table.".format(name))
                names = get_child_tables(name, engine=self.engine)
                if names:
                    return names[0]
                raise

    def reflect_table(self, name):
        if not self.has_table(name):
            raise NoSuchTableError(name)
        meta = MetaData()
        with self.engine.connect() as conn:
            table = Table(self._resolve_table_name(name), meta, autoload=True, autoload_with=conn)
            table.name = name # we do this in case we needed to use the resolved child table of a partition
            return table

    def reflect_unique_constraints(self, name):
        if not self.has_table(name):
            raise NoSuchTableError(name)
        name = self._resolve_table_name(name)
        with self.engine.connect() as conn:
            insp = inspect(conn)
            response = insp.get_pk_constraint(name)
            constraints = [KnockoffUniqueConstraint(response['constrained_columns'],
                                                    name=response['name'])]

            unique_constraints = insp.get_unique_constraints(name)

            for constraint in unique_constraints:
                constraints.append(KnockoffUniqueConstraint(constraint['column_names'],
                                                            name=constraint['name']))
            return constraints

    def insert(self, name, df, dtype=None):
        kwargs = self.kwargs.copy()
        # TODO: handle dtype[col] = JSON different?
        if dtype:
            kwargs['dtype'] = dtype
        # TODO: parallelize?
        with self.engine.connect() as conn:
            df.to_sql(name, conn,
                      **kwargs)


class KnockoffDB(object):

    def __init__(self, database_service,
                 dag_service=None,
                 seed=None):
        self.database_service = database_service
        self.dag_service = dag_service or DagService()
        self.seed = seed
        self._tables = {}

    @property
    def tables(self):
        return self._tables

    def add(self, table, insert=True, depends_on=None):
        """
        identify table by name (for depends_on?)

        build graph base on depends on
        :param table:
        :param insert:
        :param depends_on:
        :return:
        """
        # TODO: should we automatically check for KnockoffTableFactory
        #       to add thoes as dependencies?
        node = Node(table.name, table=table, insert=insert)
        self.tables[table.name] = table
        self.dag_service.add_node(node, depends_on=depends_on)

    def prepare(self):
        for node in self.dag_service.iter_topologically():
            node.table.prepare(database_service=self.database_service)


    def build(self):
        dfs = {}
        # TODO: parallelize?
        for node in self.dag_service.iter_topologically():
            table = node.table
            table.prepare(database_service=self.database_service)
            if not node.insert:
                _ = node.table.build() # create table needed downstream
            dfs[table.name] = table.df
        return dfs

    def insert(self):
        # TODO: Should we use the dfs from self.build()?
        # TODO: parallelize?
        for node in self.dag_service.iter_topologically():
            table = node.table
            table.prepare(database_service=self.database_service)
            if not node.insert:
                _ = node.table.build() # create table needed downstream
            dtype = {}
            for c, t in table.dtype.items():
                if t == dict:
                    dtype[c] = JSON
            self.database_service.insert(table.name, table.df, dtype=dtype)

    @property
    def seed(self):
        return self._seed

    @seed.setter
    def seed(self, seed):
        if seed:
            Faker.seed(seed)
            random.seed(seed)
        self._seed = seed



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


from knockoff.testing_postgresql import get_postgresql


def postgres_setup_teardown(url):
    # setup
    postgresql = get_postgresql(url=url) # create the database
    new_url = postgresql.url() # get new url

    # return the new url
    yield new_url

    # teardown
    postgresql.stop()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

from sqlalchemy import create_engine
from sqlalchemy import Table, MetaData

from ..environ import EnvironmentVariable


def get_table(name, engine=None, meta=None):
    engine = engine or get_engine()
    with engine.connect() as conn:
        meta = meta or MetaData()
        return Table(name, meta, autoload=True, autoload_with=conn)


def execute(sql, engine=None):
    engine = engine or get_engine()
    with engine.connect() as conn:
        conn.execute("commit")
        conn.execute(sql)


def get_engine(uri=None, **kwargs):
    uri = uri or get_uri()
    return create_engine(uri, **kwargs)


def get_uri(user="postgres",
            database="",
            password="password",
            host="localhost",
            port=5432,
            dialect="postgresql",
            driver=None):
    driver = '+{}'.format(driver) if driver else ''
    uri = ('{dialect}{driver}://{user}:'
           '{password}@{host}:{port}/{database}'
           ).format(dialect=dialect,
                    driver=driver,
                    user=user,
                    password=password,
                    host=host,
                    port=port,
                    database=database or '')
    return uri


class EngineBuilder(object):
    def __init__(self):
        """
        Builder for building sqlalchemy engine (or uri).

        Each component of the uri can be set with either a
        value or an environment variable. Environment variables
        will default to value passed in default parameter if
        allow_default is True. Otherwise, it will fail if
        with a KeyError if environment variable is not set.

        Components can be set multiple times, where each the
        previous set values are overridden.

        If a uri value is passed, all other components will be
        ignored. If uri environment variable is set with
        allow_default=True and default=None, uri will be used if the
        environment variable is set, otherwise it will fall back to
        the other components.

        """
        self.engine_kwargs = {}
        self.env_vars = {}

    @staticmethod
    def _assert_xor(name, value, env_var):
        assert (value is None) ^ (env_var is None), \
            ("{} must be set with either value or env (environment "
             "variable). default will be used if env_var is set "
             "and allow_default is True".format(name))

    def _remove_if_exists(self, name):
        try:
            del self.engine_kwargs[name]
        except KeyError:
            pass
        try:
            del self.env_vars[name]
        except KeyError:
            pass

    def _add_var(self, name, value, env_var,
                 default, allow_default):
        self._assert_xor(name, value, env_var)
        self._remove_if_exists(name)
        if value:
            self.engine_kwargs[name] = value
        else:
            self.env_vars[name] = \
                EnvironmentVariable(env_var,
                                    default_value=default,
                                    allow_default_value=allow_default)

    def user(self, value=None, env_var=None,
             default="postgres", allow_default=True):
        self._add_var("user", value, env_var, default, allow_default)
        return self

    def database(self, value=None, env_var=None,
                 default="", allow_default=True):
        self._add_var("database", value, env_var, default, allow_default)
        return self

    def password(self, value=None, env_var=None,
                 default="password", allow_default=True):
        self._add_var("password", value, env_var, default, allow_default)
        return self

    def host(self, value=None, env_var=None,
             default="localhost", allow_default=True):
        self._add_var("host", value, env_var, default, allow_default)
        return self

    def port(self, value=None, env_var=None,
             default=5432, allow_default=True):
        self._add_var("port", value, env_var, default, allow_default)
        return self

    def dialect(self, value=None, env_var=None,
                default="postgresql", allow_default=True):
        self._add_var("dialect", value, env_var, default, allow_default)
        return self

    def driver(self, value=None, env_var=None,
               default=None, allow_default=True):
        self._add_var("driver", value, env_var, default, allow_default)
        return self

    def uri(self, value=None, env_var=None,
            default=None, allow_default=True):
        self._add_var("uri", value, env_var, default, allow_default)
        return self

    def build(self, uri_only=False, **kwargs):
        engine_kwargs = self.engine_kwargs.copy()

        for name, variable in self.env_vars.items():
            engine_kwargs[name] = variable.get()

        try:
            uri = engine_kwargs.pop("uri")
        except KeyError:
            uri = None

        uri = uri or get_uri(**engine_kwargs)

        return uri if uri_only else get_engine(uri, **kwargs)

    @staticmethod
    def from_config(config, build='engine', **kwargs):
        """

        :param config: list
            list of dict

            e.g.

            config = [
                {
                    "property": host | port | database | user | password | dialect | driver | uri,
                    "value" : <value to set>,
                    "env_var": <environment variable to get value from>,
                    "default": <default value to use if env var isn't set>,
                    "allow_default": bool
                }
            ]

        :param build: str
            'engine', 'uri', 'builder'
        :return:
        """
        assert build in {'engine', 'uri', 'builder'}
        builder = EngineBuilder()
        for item in config:
            item = item.copy()
            name = item.pop("property")
            assert name in {"host", "port", "database",
                            "user", "password", "dialect",
                            "driver", "uri"}, \
                "property={} is not recognized".format(name)
            builder = getattr(builder, name)(**item)
        if build == "builder":
            return builder
        if build == "uri":
            return builder.build(uri_only=True)
        return builder.build(**kwargs)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import itertools


def sample_df(df, **kwargs):
    return df.sample(**kwargs).to_dict('records')[0]


def cycle_df_factory(df):
    # TODO: add parameter for number of cycles defaulting to None (inifinity)
    generator = itertools.cycle(df.iterrows())

    def _call(*args, **kwargs):
        _, row = next(generator)
        return row.to_dict()

    return _call



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import os
import pandas as pd
from sqlalchemy.pool import NullPool

from .utilities.orm.sql import EngineBuilder


KNOCKOFF_DB_URI = 'KNOCKOFF_DB_URI'
KNOCKOFF_DB_USER = 'KNOCKOFF_DB_USER'
KNOCKOFF_DB_PASSWORD = 'KNOCKOFF_DB_PASSWORD'
KNOCKOFF_DB_HOST = 'KNOCKOFF_DB_HOST'
KNOCKOFF_DB_PORT = 'KNOCKOFF_DB_PORT'
KNOCKOFF_DB_NAME = 'KNOCKOFF_DB_NAME'
KNOCKOFF_DB_DIALECT = 'KNOCKOFF_DB_DIALECT'
KNOCKOFF_DB_DRIVER = 'KNOCKOFF_DB_DRIVER'


def execute(sql, engine=None):
    engine = engine or get_engine()
    with engine.connect() as conn:
        conn.execute("commit")
        conn.execute(sql)


def create_database(database, engine=None):
    engine = engine or get_engine()
    execute("create database {};".format(database),
            engine=engine)


def create_user(user, password, engine=None):
    engine = engine or get_engine()
    execute("create user {} with encrypted password '{}';"
            .format(user, password),
            engine=engine)


DEFAULT_BUILDER = (EngineBuilder()
                   .uri(env_var=KNOCKOFF_DB_URI)
                   .host(env_var=KNOCKOFF_DB_HOST)
                   .port(env_var=KNOCKOFF_DB_PORT)
                   .user(env_var=KNOCKOFF_DB_USER)
                   .password(env_var=KNOCKOFF_DB_PASSWORD)
                   .database(env_var=KNOCKOFF_DB_NAME)
                   .dialect(env_var=KNOCKOFF_DB_DIALECT)
                   .driver(env_var=KNOCKOFF_DB_DRIVER))


ENGINE_BUILDERS = {
    "default": DEFAULT_BUILDER
}


def clear_default_env_vars():
    clear_env_vars([KNOCKOFF_DB_URI,
                    KNOCKOFF_DB_USER,
                    KNOCKOFF_DB_PASSWORD,
                    KNOCKOFF_DB_HOST,
                    KNOCKOFF_DB_PORT,
                    KNOCKOFF_DB_DIALECT,
                    KNOCKOFF_DB_DRIVER,
                    KNOCKOFF_DB_NAME,
                    KNOCKOFF_DB_URI])


def clear_env_vars(env_vars):
    # TODO: move to knockoff.utilities.environ
    for variable in env_vars:
        try:
            del os.environ[variable]
        except KeyError:
            pass


def register_engine(name, config):
    global ENGINE_BUILDERS
    ENGINE_BUILDERS[name] = EngineBuilder.from_config(config, build='builder',
                                                      poolclass=NullPool)


def reset_engines():
    global ENGINE_BUILDERS
    ENGINE_BUILDERS = {
        "default": DEFAULT_BUILDER
    }


def get_engine(name=None, build='engine', **kwargs):
    assert build in {'engine', 'builder'}
    kwargs.setdefault('poolclass', NullPool)
    name = name or "default"
    if build == 'engine':
        return ENGINE_BUILDERS[name].build(**kwargs)
    if build == 'builder':
        return ENGINE_BUILDERS[name]


def get_connection(database=None):
    engine = get_engine(database)
    return engine.connect()


def get_child_tables(partitioned_table_name, engine=None):
    """get child tables for partitioned postgres table"""
    sql = (
        "SELECT "
        "nmsp_parent.nspname AS parent_schema, "
        "parent.relname      AS parent, "
        "nmsp_child.nspname  AS child_schema, "
        "child.relname       AS child "
        "FROM pg_inherits "
        "JOIN pg_class parent            ON pg_inherits.inhparent = parent.oid "
        "JOIN pg_class child             ON pg_inherits.inhrelid   = child.oid "
        "JOIN pg_namespace nmsp_parent   ON nmsp_parent.oid  = parent.relnamespace "
        "JOIN pg_namespace nmsp_child    ON nmsp_child.oid   = child.relnamespace "
        "WHERE parent.relname='{table}';"
    ).format(table=partitioned_table_name)
    engine = engine or get_engine()
    with engine.connect() as conn:
        df = pd.read_sql_query(sql, conn)
    return df["child"].tolist()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

from dependency_injector import containers, providers
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool

from knockoff.sdk.db import KnockoffDB, DefaultDatabaseService
from knockoff.sdk.blueprint import Blueprint

class KnockoffContainer(containers.DeclarativeContainer):
    config = providers.Configuration()

    engine = providers.Factory(
        create_engine,
        url=config.database_service.url,
        poolclass=NullPool
    )

    database_service = providers.Singleton(
        DefaultDatabaseService,
        engine=engine
    )

    knockoff_db = providers.Singleton(
        KnockoffDB,
        database_service=database_service
    )

    blueprint = providers.Factory(
        Blueprint.from_plan_package_name,
        config.blueprint.plan.package)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


from faker import Faker
from numpy import random

import yaml

from . import orm
from .factory.assembly import Blueprint, Assembler
from .factory.node import Table, FactoryPart
from .factory.node import FactoryComponent, FactoryPrototype


class Knockoff(object):

    def __init__(self, path_or_config, seed=None):

        if isinstance(path_or_config, dict):
            self.config = path_or_config
        else:
            with open(path_or_config, "rb") as f:
                self.config = yaml.safe_load(f)

        self.seed = seed

        self.__register_database_engines()

        blueprint = Blueprint.from_config(self.config["knockoff"])
        assembler = Assembler(blueprint)
        assembler.start()

        self.__register_nodes(blueprint)

    def __register_database_engines(self):
        for config in self.config["knockoff"].get("databases", []):
            orm.register_engine(config["name"], config["config"])

    def __register_nodes(self, blueprint):
        data = {}
        for node in blueprint.nodes:
            if node.node_type in {Table.node_type,
                                  FactoryPart.node_type,
                                  FactoryPrototype.node_type}:
                data.setdefault(node.node_type, {})[node.name] = node.data
            elif node.node_type == FactoryComponent.node_type:
                data.setdefault(node.node_type, {})[node.name] = node.generator
        self.tables = data.get(Table.node_type, {})
        self.parts = data.get(FactoryPart.node_type, {})
        self.prototypes = data.get(FactoryPrototype.node_type, {})
        self.components = data.get(FactoryComponent.node_type, {})

    @property
    def seed(self):
        return self._seed

    @seed.setter
    def seed(self, seed):
        if seed:
            Faker.seed(seed)
            random.seed(seed)
        self._seed = seed



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

def load_knockoff(source, assembler, node_name):
    kwargs = source.config["kwargs"]
    df = assembler.blueprint.get_node("prototype",
                                      kwargs["prototype"]).data
    columns = kwargs.get('columns')
    if columns:
        return df[columns]
    return df



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import os


class EnvironmentVariable(object):
    def __init__(self, variable,
                 default_value=None,
                 allow_default_value=False):
        self.variable = variable
        self.default_value = default_value
        self.allow_default_value = allow_default_value

    def get(self):
        if self.allow_default_value:
            return os.getenv(self.variable, self.default_value)
        return os.environ[self.variable]

    def set(self, value):
        os.environ[self.variable] = value



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import argparse
import inspect
import sys
import logging

from dependency_injector import containers

from knockoff.utilities.importlib_utils import resolve_package_name
from knockoff.sdk.db import KnockoffDB, DefaultDatabaseService


logger = logging.getLogger(__name__)


def run(knockoff_db: KnockoffDB,
        blueprint):
    dfs, knockoff_db = blueprint.construct(knockoff_db)
    knockoff_db.insert()
    logger.info("knockoff done.")


def _validate_container_class(cls, package_name):
    if not inspect.isclass(cls) or not issubclass(cls, containers.DeclarativeContainer):
        raise TypeError(f"{package_name} resolves to "
                         f"{cls} instead of "
                         f"a subclass of dependency_injector"
                         f".containers.DeclarativeContainer")


def seed(i):
    """TODO: move this to a utility module?"""
    import random
    import numpy as np
    from faker import Faker
    Faker.seed(i)
    np.random.seed(i)
    random.seed(i)
    logger.info(f"Seeds for Faker, Numpy, and random "
                f"modules have been set to {i}")


def parse_args(argv=None):
    parser = argparse.ArgumentParser(
        usage='''knockoff run [<args>]'''
    )
    parser.add_argument("-c", "--container",
                        default="knockoff.sdk.container.default:KnockoffContainer",
                        help="Default KnockoffContainer")
    parser.add_argument("--yaml-config",
                        help="Container configuration")
    parser.add_argument("-s", "--seed", type=int,
                        help="Set seed")
    return parser.parse_args(argv)


def main(argv=None):
    """
    TODO:
            - Add ephemeral flag for spinning up temporary database that's destroyed
              after program completion (with interactive program)
    """
    args = parse_args(argv)

    KnockoffContainer = resolve_package_name(args.container)
    _validate_container_class(KnockoffContainer, args.container)
    container = KnockoffContainer()
    container.init_resources()
    if args.yaml_config:
        container.config.from_yaml(args.yaml_config)
    container.wire(modules=[sys.modules[__name__]])

    if args.seed:
        seed(args.seed)

    knockoff_db = container.knockoff_db()
    blueprint = container.blueprint()
    run(knockoff_db, blueprint)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

from .setup_teardown import postgres_setup_teardown


class TempDatabaseService(object):
    def __init__(self,
                 url,
                 setup_teardown=postgres_setup_teardown,
                 initialize_tables=None):
        self.url = url
        self.setup_teardown = setup_teardown
        self.initialize_tables = initialize_tables
        self.temp_url = None
        self._setup_teardown_generator = None

    def start(self):
        if self._setup_teardown_generator is not None:
            raise ValueError(f"Invalid state reached on start()."
                             f"Possible causes: \n1) start() has "
                             f"already been called. \n2) Bad "
                             f"generator function returned by"
                             f" {self.setup_teardown}")
        self._setup_teardown_generator = self.setup_teardown(self.url)
        try:
            self.temp_url = next(self._setup_teardown_generator)
        except StopIteration:
            raise ValueError(f"Invalid state reached on start()."
                             f"Possible causes: \n1) Bad "
                             f"generator function returned by"
                             f" {self.setup_teardown}")
        except TypeError:
            raise ValueError(f"{self.setup_teardown} did not return a generator")

        if self.initialize_tables:
            self.initialize_tables(self.temp_url)
        return self.temp_url

    def stop(self):
        if self._setup_teardown_generator is None:
            raise ValueError(f"Invalid state reached on stop()."
                             f"Possible causes: \n1) start() was not called."
                             f"\n2) stop() was already called.")
        try:
            next(self._setup_teardown_generator)
        except StopIteration:
            return
        raise ValueError(f"Invalid state reached on stop()."
                         f"Possible causes: \n1) Bad "
                         f"generator function returned by"
                         f" {self.setup_teardown}")

    def reset(self):
        self.temp_url = None
        self._setup_teardown_generator = None



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import os
import re
import testing.postgresql
from uuid import uuid4
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool


TRUTH_VALUES =  {'1', 't', 'true'}

TEST_POSTGRES_ENABLED = os.getenv("TEST_POSTGRES_ENABLED",
                                  '1').lower() in TRUTH_VALUES

TEST_USE_EXTERNAL_DB = os.getenv('TEST_USE_EXTERNAL_DB',
                                 '1').lower() in TRUTH_VALUES


DEFAULT_URL = "postgresql://postgres@localhost:5432/postgres"


class ExternalPostgresql(object):
    def __init__(self, url=DEFAULT_URL,
                 db_name=None):

        self.db_name = db_name or 'test_' + str(uuid4()).replace('-', '')

        self.engine = create_engine(url)
        with self.engine.connect() as conn:
            conn.execute("commit")
            conn.execute("CREATE DATABASE {};".format(self.db_name))

    def url(self):
        # Substitute db name in URI with generated db
        url = re.sub(
            "(?:/[A-Za-z0-9_]+)$", "/{}".format(self.db_name), str(self.engine.url)
        )
        engine = create_engine(url, pool=NullPool)
        url = engine.url
        url = str(url)
        return url

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.stop()

    def stop(self):
        terminate_connections_sql = (
            "SELECT pg_terminate_backend(pid) "
            "FROM pg_stat_activity "
            "WHERE datname ='{}'"
        ).format(self.db_name)

        with self.engine.connect() as conn:
            conn.execute("commit")
            conn.execute(terminate_connections_sql)
            conn.execute("DROP DATABASE {};".format(self.db_name))
        self.engine.dispose()


def get_postgresql(url=DEFAULT_URL):

    if TEST_USE_EXTERNAL_DB:
        return ExternalPostgresql(url=url)

    return testing.postgresql.Postgresql()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import pkg_resources

__version__ = pkg_resources.get_distribution(__name__).version




# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import six
from abc import ABCMeta, abstractmethod

from .functools import abstractproperty2to3


class InvalidStringError(ValueError):
    """ Raised when an invalid string is encountered """

    def __init__(self, string, regex):
        self.string = string
        self.regex = regex

    def __str__(self):
        return ('Invalid string: {}. Regex: {}'
                .format(repr(self.string),
                        repr(self.regex)))


class RegexParser(six.with_metaclass(ABCMeta, object)):

    @abstractproperty2to3
    def compiled_regex(self):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def make(cls, **kwargs):
        return

    @classmethod
    def parse(cls, string):
        m = cls.compiled_regex.match(string)
        if not m:
            raise InvalidStringError(string,
                                     cls.compiled_regex.pattern)
        kwargs = m.groupdict()
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return cls.make(**kwargs)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import pandas as pd
from .next_strategy.df import sample_df
from .next_strategy.table import sample_table


class KnockoffFactory(object):
    def __init__(self, obj, columns=None, rename=None, drop=None,
                 next_strategy_callable=None,
                 next_strategy_factory=None,
                 lazy_init=True):

        self.obj = obj
        self.columns = columns
        self.rename = rename or {}
        self.drop = set(drop or [])
        self.lazy_init = lazy_init
        self.next = None
        self.initialized = False
        assert (next_strategy_callable is None) ^ (next_strategy_factory is None)
        self.next_strategy_callable = next_strategy_callable
        self.next_strategy_factory = next_strategy_factory
        if not self.lazy_init:
            self.initialize()

    def initialize(self):
        if self.next_strategy_factory:
            self.next = self.next_strategy_factory(self.obj)
        else:
            self.next = self.next_strategy_callable
        self.initialized = True

    def __call__(self):
        if not self.initialized:
            self.initialize()
        record = self.next(self.obj)
        out = {}
        for col in self.columns:
            if col in self.drop:
                continue
            out[self.rename.get(col, col)] = record[col]
        return out


class KnockoffTableFactory(KnockoffFactory):

    def __init__(self, table, columns=None, rename=None, drop=None,
                 next_strategy_callable=sample_table,
                 next_strategy_factory=None):
        if (next_strategy_factory is None) and (next_strategy_callable is None):
            next_strategy_callable = sample_table
        super(KnockoffTableFactory, self).__init__(table,
                                                   columns=columns,
                                                   rename=rename,
                                                   drop=drop,
                                                   next_strategy_callable=next_strategy_callable,
                                                   next_strategy_factory=next_strategy_factory)

    def initialize(self):
        super(KnockoffTableFactory, self).initialize()
        if self.columns is None:
            self.columns = self.columns or self.obj.columns


class KnockoffDataFrameFactory(KnockoffFactory):

    def __init__(self, df, columns=None, rename=None, drop=None,
                 next_strategy_callable=None,
                 next_strategy_factory=None):
        if (next_strategy_factory is None) and (next_strategy_callable is None):
            next_strategy_callable = sample_df
        assert isinstance(df, pd.DataFrame)
        super(KnockoffDataFrameFactory, self).__init__(df,
                                                       columns=columns,
                                                       rename=rename,
                                                       drop=drop,
                                                       next_strategy_callable=next_strategy_callable,
                                                       next_strategy_factory=next_strategy_factory)

    def initialize(self):
        super(KnockoffDataFrameFactory, self).initialize()
        if self.columns is None:
            self.columns = self.columns or self.obj.columns


class KnockoffTransform(object):
    def __init__(self, factory, transform):
        self.factory = factory
        self.transform = transform

    def __call__(self):
        return self.transform(self.factory())



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import networkx as nx
from networkx.algorithms import topological_sort
from networkx.algorithms import is_directed_acyclic_graph


class Node(object):
    def __init__(self, node_id, **kwargs):
        self._node_id = node_id
        for key, value in kwargs.items():
            assert key not in {'node_id', '_node_id'}
            setattr(self, key, value)

    @property
    def node_id(self):
        return self._node_id


class DagService(object):
    def __init__(self):
        self.dag = nx.DiGraph()
        self.nodes = []
        self.index = {}

    def get_node(self, node_id):
        return self.nodes[self.index[node_id]]

    def add_node(self, node, depends_on=None):
        node_id = node.node_id
        assert node_id not in self.index, 'Node already present'
        ix = len(self.nodes)
        self.nodes.append(node)
        self.index[node_id] = ix
        self.dag.add_node(node_id)
        if depends_on:
            self.dag.add_edges_from([(other_node_id, node_id) for other_node_id in depends_on])

    def iter_topologically(self):
        assert is_directed_acyclic_graph(self.dag)
        return (self.nodes[self.index[node_id]] for node_id in topological_sort(self.dag))



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


from .date.interval import Interval


def generate_periods_from_intervals(bop_interval, eop_interval):
    """

    :param bop_interval: beginning of period interval
    :param eop_interval: end of period interval
    :return: list of tuple of (bop, eop) date objects
    """
    bop_dates = Interval.parse(bop_interval)
    eop_dates = Interval.parse(eop_interval)

    assert len(bop_dates) == len(eop_dates), \
        "Length mismatch: %s!=%s. bop and eop intervals must generate lists of" \
        " same length" % (len(bop_dates), len(eop_dates))

    return [
        (bop, eop) for bop, eop in zip(bop_dates, eop_dates)
    ]



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


from importlib import import_module


def resolve_package_name(package_name):
    """
    Parameters
    ----------

    package_name: str
        The expected format of package_name is 'package:name'
        where 'package' should be an importable module name and
        'name' should be the name of an object accessible within
        that module
    """
    package, name = package_name.split(':')
    module = import_module(package)
    return getattr(module, name)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

from knockoff.utilities.mixin import ResourceLocatorMixin


class ComponentFunctionFactory(ResourceLocatorMixin, object):
    entry_point_group = "knockoff.factory.component.function"


class ComponentMixin:
    source = None
    generator = None
    name = None

    def prepare(self, assembler):
        if self.name is None:
            raise Exception("Missing name")
        if self.source is None:
            raise Exception('Missing source')
        if self.generator is None:
            self.generator = self.source.load(assembler, self.name)


def load_autoincrement(source, assembler, name):

    def autoincrement_generator():
        i = source.config.get('start_value', 0)
        while True:
            yield i
            i += 1

    return autoincrement_generator()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import os
import six
import logging
import itertools
from operator import itemgetter

import pandas as pd

from knockoff.factory.component import ComponentFunctionFactory
from knockoff.factory.counterfeit import KNOCKOFF_ATTEMPT_LIMIT_ENV
from knockoff.utilities.functools import call_with_args_kwargs


logger = logging.getLogger(__name__)


def load_prototype_from_components(source, assembler, node_name):
    limit = int(os.environ.get(KNOCKOFF_ATTEMPT_LIMIT_ENV, 1000000))

    dependencies = [(component['name'],
                     dep) for component in source.config['components']
                    for dep in component['source'].get('dependencies',
                                                       [])]

    def sort(name_dep):
        return tuple(assembler.parse_dependency(name_dep[1])[:-1])

    dependencies.sort(key=sort)
    names = [component['name'] for component in source.config['components']]
    name_to_source = {component['name']:component['source']
                      for component in source.config['components']}

    i = 0
    records = []
    # create mapping of unique key indices to unique keys
    unique_constraints = {tuple(constraint): set()
                          for constraint in (source.config.get('unique',
                                                               []))}
    while len(records) < source.config['number'] and i < limit:
        i += 1
        record = {}
        component_to_function_input = {}
        for (node_type, name), group in itertools.groupby(dependencies,
                                                          sort):
            node = assembler.blueprint.get_node(node_type, name)
            sample = node.data.sample(1)
            for component_name, dep in group:
                strategy = name_to_source[component_name]['strategy']
                if strategy == "knockoff":
                    (dep_type,
                     dep_name,
                     dep_sub_name) = assembler.parse_dependency(dep)
                    col = dep_sub_name or 0
                    record[component_name] = sample[col].values[0]
                elif strategy == "function":
                    (dep_type,
                     dep_name,
                     dep_sub_name) = assembler.parse_dependency(dep)
                    col = dep_sub_name or 0
                    (component_to_function_input
                     .setdefault(component_name,
                                 {}))[dep] = sample[col].values[0]
                else:
                    raise Exception("strategy not recognized: {}"
                                    .format(strategy))

        function_factory = ComponentFunctionFactory()

        for component_name, source_config in six.iteritems(name_to_source):
            if component_name in record:
                continue
            strategy = source_config['strategy']
            if strategy == 'function':
                func = function_factory.get_resource(source_config['function'])
                record[component_name] = \
                    handle_function(func,
                                    input_args=(source_config
                                                .get('input_args')),
                                    input_kwargs=(source_config
                                                  .get('input_kwargs')),
                                    func_inputs_from_dependencies=
                                    component_to_function_input
                                    .get(component_name))

            else:
                node = assembler.blueprint.get_node("component", "{}.{}"
                                                    .format(node_name,
                                                            component_name))
                record[component_name] = next(node.generator)

        if not _satisfies_unique_constraints(record, unique_constraints):
            continue

        for constraint in six.iterkeys(unique_constraints):
            unique_constraints[constraint].add(itemgetter(*constraint)(record))

        records.append(itemgetter(*names)(record))

    if i >= limit:
        logger.error("Could not generate prototype={}"
                     .format(node_name))
        raise Exception("Attempts to create unique set reached: {}"
                        .format(limit))
    return pd.DataFrame(records, columns=names)


def handle_function(func, input_args=None, input_kwargs=None,
                    func_inputs_from_dependencies=None):

    args = []
    kwargs = {}

    for input_cfg in input_args or []:
        if input_cfg['type'] == "constant":
            args.append(input_cfg['value'])
        elif input_cfg['type'] == 'dependency':
            args.append(func_inputs_from_dependencies[input_cfg['value']])
    for input_cfg in input_kwargs or []:
        if input_cfg['type'] == "constant":
            kwargs[input_cfg['key']] = input_cfg['value']
        elif input_cfg['type'] == 'dependency':
            kwargs[input_cfg['key']] = \
                func_inputs_from_dependencies[input_cfg['value']]
    return call_with_args_kwargs(func, tuple(args), kwargs)


def _satisfies_unique_constraints(record, constraints):
    valid_record = True
    for constraint in six.iterkeys(constraints):
        if itemgetter(*constraint)(record) in constraints[constraint]:
            valid_record = False
        if not valid_record:
            break
    return valid_record



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import os
import sys
import pipes
import argparse
import yaml
import subprocess
import logging

import sqlalchemy_utils
from sqlalchemy import create_engine
from sqlalchemy.exc import ProgrammingError

from knockoff.utilities.functools import call_with_args_kwargs
from .utilities.orm.sql import EngineBuilder

from . import orm, __version__
from .io import ReaderFactory
from .knockoff import Knockoff
from .cli_v2 import setup_logger


logger = logging.getLogger(__name__)


def shlex_join(cmd):
    return " ".join(map(pipes.quote, cmd))


def _add_flag_from_env(env_var, flag, cmd):
    if env_var in os.environ:
        cmd.append(flag)
        cmd.append(os.environ[env_var])


def assert_supported_db_action(db_type):
    assert db_type in {'postgres'}, ("database.type={} "
                                     "is not supported"
                                     "for this operation."
                                     .format(db_type))


def create_databases(create_database_configs):
    for config in create_database_configs:
        db_type = config.get('type')
        engine_config = config.get('config')
        if engine_config:
            engine = EngineBuilder.from_config(engine_config)
        else:
            engine = orm.get_engine()
        assert_supported_db_action(db_type)

        parts = str(engine.url).split('/')
        url = '/'.join(parts[:-1] + [config['name']])

        if sqlalchemy_utils.database_exists(url):
            logger.info("{} already exists. Skipping creation."
                        .format(config["name"]))
        else:
            # the following two lines remove configuration to specific db
            url = '/'.join(parts[:-1])
            engine = create_engine(url)
            orm.create_database(config['name'], engine=engine)
        for user_config in config.get('users', []):
            assert_supported_db_action(db_type)
            password = os.environ[user_config['password_env']]
            try:
                orm.create_user(user_config['user'], password,
                                engine=engine)
                orm.execute("grant all privileges on database {} to {};"
                            .format(config['name'], user_config['user']),
                            engine=engine)
            except ProgrammingError as e:
                logger.error(e)
                logger.warning("Error creating user. Most likely exists."
                               "Skipping this step.")


def load_schemas(yamltodb_configs):
    for config in yamltodb_configs:
        logger.info("Loading tables from {} to database: {}"
                    .format(config['path'], config['database']))
        cmd = []
        if orm.KNOCKOFF_DB_PASSWORD in os.environ:
            cmd.append("{}={}".format('PGPASSWORD',
                                      os.environ[orm
                                                 .KNOCKOFF_DB_PASSWORD]))
        cmd.append("yamltodb")
        _add_flag_from_env(orm.KNOCKOFF_DB_HOST, "-H", cmd)
        _add_flag_from_env(orm.KNOCKOFF_DB_USER, "-U", cmd)
        cmd.append(config['database'])
        cmd.append(config['path'])
        cmd.append("--update")
        subprocess.check_call(shlex_join(cmd), shell=True)
        logger.info("Loaded tables from {} to database: {}"
                    .format(config['path'], config['database']))


def load_df_from_source(source):
    reader = ReaderFactory().get_resource(source['reader'])
    args = source.get('args')
    kwargs = source.get('kwargs')

    return call_with_args_kwargs(reader, args, kwargs)


def parse_args(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config",
                        default=os.environ.get('KNOCKOFF_CONFIG',
                                               'knockoff.yaml'),
                        help="path to knockoff config")
    parser.add_argument("-s", "--seed",
                        type=int,
                        default=os.environ.get('KNOCKOFF_SEED'),
                        help="path to knockoff config")
    parser.add_argument("-v", "--verbose",
                        action="store_true",
                        help="verbose logging")
    parser.add_argument("--version",
                        action="store_true",
                        help="print knockoff version")
    args = parser.parse_args(argv)
    return args


def main(argv=None):
    args = parse_args(argv)

    if args.version:
        print(__version__)
        sys.exit(0)

    setup_logger(args.verbose)

    with open(args.config, 'rb') as configfile:
        config = yaml.safe_load(configfile)

    create_databases(config.get('create-databases', []))

    if 'yamltodb' in config:
        load_schemas(config['yamltodb'])

    Knockoff(config, seed=args.seed)


if __name__ == "__main__":
    main()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import sys
import logging
import glob

import pandas as pd
import s3fs
from pyarrow import parquet as pq

from .utilities.functools import call_with_args_kwargs
from .utilities.mixin import ResourceLocatorMixin
from .orm import get_connection

if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO

logger = logging.getLogger(__name__)


class ReaderFactory(ResourceLocatorMixin, object):
    entry_point_group = "knockoff.io.readers"


class WriterFactory(ResourceLocatorMixin, object):
    entry_point_group = "knockoff.io.writers"


def load_strategy_io(source, assembler, name):
    reader = ReaderFactory().get_resource(source.config['reader'])
    args = source.config.get('args')
    kwargs = source.config.get('kwargs')

    return call_with_args_kwargs(reader, args, kwargs)


def read_inline(data, **kwargs):
    return pd.read_csv(StringIO(data), **kwargs)


def read_multi_parquet(path):
    if path.startswith("s3://"):
        s3 = s3fs.S3FileSystem()
        fs = s3fs.core.S3FileSystem()

        s3_path = path.replace("s3://", "", 1)
        s3_paths = fs.glob(path=s3_path)
        return pq.ParquetDataset(s3_paths, filesystem=s3).read_pandas().to_pandas()

    paths = glob.glob(path)
    return pq.ParquetDataset(paths).read_pandas().to_pandas()


def read_sql(sql, database, **kwargs):
    with get_connection(database) as conn:
        return pd.read_sql_query(sql, conn, **kwargs)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import six
from abc import ABCMeta, abstractmethod
from operator import itemgetter


class KnockoffConstraint(six.with_metaclass(ABCMeta, object)):
    @abstractmethod
    def reset(self):
        """reset to initial state"""
        return

    @abstractmethod
    def check(self, record):
        """if record would satisfy constraint return True"""
        return

    def add(self, record):
        if not self.check(record):
            raise ValueError('Should raise a constraint error here')


class KnockoffUniqueConstraint(KnockoffConstraint):
    def __init__(self, keys, name=None):
        self.name = name # is this necessary?
        assert len(keys) > 0
        self.keys = keys
        self.curr_set = set()
        self.parse = itemgetter(*self.keys)

    def check(self, record):
        return self.parse(record) not in self.curr_set

    def add(self, record):
        # TODO: do we care if it already exists?
        self.curr_set.add(self.parse(record))

    def reset(self):
        self.curr_set = set()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import sys

from knockoff import __version__


def main(argv=None):
    print(f"knockoff {__version__}")


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


from sqlalchemy.orm import sessionmaker


class SqlAlchemyConn(object):
    def __init__(self, engine):
        self.engine = engine
        Session = sessionmaker(bind=self.engine)
        self.session = Session()


    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if exception_type:
            self.session.rollback()
        else:
            self.session.commit()
        self.session.close()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

from collections import namedtuple

from knockoff.utilities.importlib_utils import resolve_package_name

class Blueprint(object):

    ConstructionResult = namedtuple('ConstructionResult', [
        'dfs',
        'knockoff_db'
    ])

    def __init__(self, plan):
        """
        Parameters
        ----------
        plan: function
            plan(knockoff_db: KnockoffDB) -> KnockoffDB
        """

        self.plan = plan

    @classmethod
    def from_plan_package_name(cls, package_name):
        plan = resolve_package_name(package_name)
        return Blueprint(plan)

    def construct(self, knockoff_db):
        knockoff_db = self.plan(knockoff_db)
        dfs = knockoff_db.build()
        return Blueprint.ConstructionResult(dfs=dfs,
                                            knockoff_db=knockoff_db)


def noplan(knockoff_db):
    """no plan used for testing"""
    return knockoff_db



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


from sqlalchemy import create_engine


class SqlAlchemyInitTablesFunc(object):
    def __init__(self, base):
        """
        Parameters
        ----------
        base: sqlachlemy.ext.declarative_base() class
        """
        self.base = base

    def __call__(self, url):
        engine = create_engine(url)
        self.base.metadata.create_all(engine)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import pandas as pd

from ..utilities.period import generate_periods_from_intervals
from datetime import datetime


def concat_strategy(source, assembler, node_name):
    dfs = []
    for dependency in source.config['dependencies']:
        node_type, name, sub_name = assembler.parse_dependency(dependency)
        node = assembler.blueprint.get_node(node_type, name)
        if sub_name:
            dfs.append(node.data[[sub_name]])
        else:
            dfs.append(node.data)
    return pd.concat(dfs, axis=0)


def cartesian_product_strategy(source, assembler, node_name):
    values = []
    assert (len(source.config['dependencies'])
            == len(source.config['index'])), ('Number of elements in '
                                              'dependencies and index '
                                              'must be equal.')
    for dependency in source.config['dependencies']:
        node_type, name, sub_name = assembler.parse_dependency(dependency)
        node = assembler.blueprint.get_node(node_type, name)
        if sub_name:
            values.append(node.data[[sub_name]].values)
        else:
            values.append(node.data[0].values)
    index = pd.MultiIndex.from_product(values, names=source.config['index'])
    return pd.DataFrame(index=index).reset_index()


def read_part_inline(source, assembler, node_name):
    return pd.DataFrame(source.config['data'])


def generate_part_periods(source, assembler, node_name):

    periods = generate_periods_from_intervals(
        bop_interval=source.config['bop_interval'],
        eop_interval=source.config['eop_interval'],
    )

    string_format = source.config.get("string_format")

    if string_format:
        def to_string(x):
            return (
                datetime.strftime(x[0], string_format),
                datetime.strftime(x[1], string_format)
            )

        periods = [to_string(x) for x in periods]

    return pd.DataFrame(list(zip(periods)))



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

from knockoff.utilities.mixin import ResourceLocatorMixin


class SinkDumpStrategyFactory(ResourceLocatorMixin, object):
    entry_point_group = "knockoff.factory.sink.dump_strategy"


class KnockoffSink:
    def __init__(self, config):
        self.config = config
        self.dump_strategy = (SinkDumpStrategyFactory()
                              .get_resource(config.get('strategy', 'noop')))

    def dump(self, table, df):
        return self.dump_strategy(self, table, df)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import re
import operator
from datetime import datetime, date

import dateutil
import dateutil.parser
from dateutil.relativedelta import SU, MO, TU, WE, TH, FR, SA
from dateutil.relativedelta import relativedelta

from ..regex import RegexParser


class UnitSplitter(RegexParser):

    compiled_regex = re.compile(
        r'(?P<multiple>[-+]?[0-9]*)?(?P<unit>[a-zA-Z]+)'
    )

    @classmethod
    def make(cls, unit, multiple=None):
        if not multiple or multiple == '+':
            multiple = 1
        elif multiple == '-':
            multiple = -1
        else:
            multiple = int(multiple)
        return multiple, unit


# Units mapped to parameter in relativedelta
UNIT = {
    'y': 'years',
    'm': 'months',
    'w': 'weeks',
    'd': 'days',
    'H': 'hours',
    'M': 'minutes',
    's': 'seconds',
}

# Only allow units we know how to map
UNITS_RGX = r'[{}]'.format(''.join(UNIT.keys()))

WEEKDAY = {
    'SU': SU,
    'MO': MO,
    'TU': TU,
    'WE': WE,
    'TH': TH,
    'FR': FR,
    'SA': SA,
}


def now(utc=False):
    """Make easier to unittest"""
    return datetime.utcnow() if utc else datetime.now()


ISO8601_RGX = (r'(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])'
               r'-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):'
               r'([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-]'
               r'(?:2[0-3]|[01][0-9]):[0-5][0-9])?')
DATE_RGX = (r'(?:[12]\d{3}-(?:0[1-9]|1[0-2])'
            r'-(?:0[1-9]|[12]\d|3[01]))')

INTERVAL_DATE_RGX = (r'(?:today|now|TODAY|NOW|utctoday|utcnow|UTCTODAY|UTCNOW|{}|{})'
                     .format(ISO8601_RGX, DATE_RGX))
RDELTA_RGX = r'[-+]\d*{}'.format(UNITS_RGX)
WEEKDAY_RGX = r'[-+]\d*(?:MO|TU|WE|TH|FR|SA|SU)'


class Endpoint(RegexParser):

    compiled_regex = re.compile(
        (
            r'^(?P<reference>{})?'
            r'\s*(?P<rdelta>{})?\s*'
            r'\s*(?P<weekday>{})?\s*$'
        ).format(INTERVAL_DATE_RGX, RDELTA_RGX, WEEKDAY_RGX)
    )

    @classmethod
    def make(cls, reference=None, rdelta=None, weekday=None):
        kwargs = {}
        if rdelta:
            multiple, unit = UnitSplitter.parse(rdelta)
            kwargs[UNIT[unit]] = multiple
        if weekday:
            multiple, weekday = UnitSplitter.parse(weekday)
        if reference:
            reference = cls.parse_reference(reference)
        if rdelta and weekday:
            rdelta = relativedelta(weekday=WEEKDAY[weekday](multiple),
                                   **kwargs)
        elif rdelta:
            rdelta = relativedelta(**kwargs)
        elif weekday:
            rdelta = relativedelta(weekday=WEEKDAY[weekday](multiple))
        if rdelta and reference:
            return reference + rdelta
        return reference if reference else rdelta

    @staticmethod
    def parse_reference(reference_string):
        reference_string_lower = reference_string.lower()
        if reference_string_lower == 'today':
            return now().date()
        elif reference_string_lower == 'now':
            return now()
        elif reference_string_lower == 'utctoday':
            return now(utc=True).date()
        elif reference_string_lower == 'utcnow':
            return now(utc=True)
        return dateutil.parser.parse(reference_string)


class IntervalFrequency(RegexParser):

    regex = r'\d*{}'.format(UNITS_RGX)

    compiled_regex = re.compile(
        r'^(?:\s*(?P<interval_freq>{})\s*)$'.format(regex)
    )

    @classmethod
    def make(cls, interval_freq):
        multiple, unit = UnitSplitter.parse(interval_freq)
        return relativedelta(**{UNIT[unit]: int(multiple)})


class Interval(RegexParser):

    compiled_regex = re.compile(
        (
            r'^(?P<left>[\[(])\s*(?P<start>{0})'
            r'\s*,\s*'
            r'(?P<end>{0})\s*(?P<right>[)\]])'
            r'(?:\s*\|\s*(?P<interval_freq>{1})\s*)?$'
        ).format(r'.+', IntervalFrequency.regex)
    )

    @staticmethod
    def divisible_by_day(dt):
        seconds = (dt - datetime(1970, 1, 1)).total_seconds()
        return int(seconds) % 86400 == 0

    @staticmethod
    def _should_use_date(dates):
        is_first = (type(dates[0]) is date
                    or Interval.divisible_by_day(dates[0]))
        if len(dates) == 1:
            return is_first
        return is_first and (type(dates[1]) is date
                             or Interval.divisible_by_day(dates[1]))

    @staticmethod
    def to_datetime(dt):
        if type(dt) is datetime:
            return dt
        elif type(dt) is date:
            return datetime.fromordinal(dt.toordinal())
        else:
            raise ValueError('Unrecognized type: {}'.format(type(dt)))

    @staticmethod
    def to_date(dt):
        if type(dt) is datetime:
            return dt.date()
        elif type(dt) is date:
            return dt
        else:
            raise ValueError('Unrecognized type: {}'.format(type(dt)))

    @staticmethod
    def resolve_range(start, end):
        start = Endpoint.parse(start)
        end = Endpoint.parse(end)
        is_start_rel = isinstance(start, relativedelta)
        is_end_rel = isinstance(end, relativedelta)
        assert not (is_start_rel and is_end_rel)
        if not is_start_rel and not is_end_rel:
            return start, end
        if is_start_rel:
            return start + end, end
        return start, start + end

    @classmethod
    def make(cls, start, end, left='[', right=')', interval_freq='d'):

        start, end = cls.resolve_range(start, end)

        interval_freq = IntervalFrequency.parse(interval_freq)
        left_closed = left == '['
        right_closed = right == ']'

        # initialize as datetime for initial operators
        start = cls.to_datetime(start)
        end = cls.to_datetime(end)

        if start < end:
            increment = operator.add
            compare_open = operator.lt
            compare_closed = operator.le
        else:
            increment = operator.sub
            compare_open = operator.gt
            compare_closed = operator.ge

        cur = increment(start, interval_freq)

        if cls._should_use_date([start, cur]):
            start = cls.to_date(start)
            cur = cls.to_date(cur)
            end = cls.to_date(end)
        else:
            start = cls.to_datetime(start)
            cur = cls.to_datetime(cur)
            end = cls.to_datetime(end)

        dates = []

        if left_closed:
            dates.append(start)

        while ((compare_open(cur, end)
                and not right_closed)
               or (compare_closed(cur, end)
                   and right_closed)):
            dates.append(cur)
            cur = increment(cur, interval_freq)

        return dates



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import sys
from abc import abstractmethod, abstractproperty


def call_with_args_kwargs(function, args, kwargs):
    has_args = args is not None and len(args) > 0
    has_kwargs = kwargs is not None and len(kwargs) > 0
    if has_args and has_kwargs:
        return function(*args, **kwargs)
    elif has_args:
        return function(*args)
    elif has_kwargs:
        return function(**kwargs)
    return function()


def abstractproperty2to3(func):

    if sys.version_info > (3, 3):
        return property(abstractmethod(func))
    else:
        return abstractproperty(func)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import six
import logging
from abc import ABCMeta, abstractmethod

from knockoff.factory.source import SourceMixin
from knockoff.factory.component import ComponentMixin


logger = logging.getLogger(__name__)


class KnockoffNode(six.with_metaclass(ABCMeta, object)):

    node_type = None

    def __init__(self, name, ix):
        self.name = name
        self.ix = ix

    @abstractmethod
    def visit(self):
        raise NotImplementedError


class Table(SourceMixin, KnockoffNode):

    node_type = "table"

    def __init__(self, name, ix, source, sink, table=None):
        super(Table, self).__init__(name, ix)
        self.table = table or name
        self.source = source
        self.sink = sink

    def visit(self, assembler):
        assembler.assemble_table(self)


class FactoryPart(SourceMixin, KnockoffNode):

    node_type = "part"

    def __init__(self, name, ix, source):
        super(FactoryPart, self).__init__(name, ix)
        self.source = source

    def visit(self, assembler):
        assembler.assemble_part(self)


class FactoryComponent(ComponentMixin, KnockoffNode):

    node_type = "component"

    def __init__(self, name, ix, source):
        super(FactoryComponent, self).__init__(name, ix)
        self.source = source

    def visit(self, assembler):
        assembler.prepare_component(self)


class FactoryPrototype(SourceMixin, KnockoffNode):

    node_type = "prototype"

    def __init__(self, name, ix, source, unique=None):
        super(FactoryPrototype, self).__init__(name, ix)
        self.source = source
        self.unique = unique or []
        self.data = None

    def visit(self, assembler):
        assembler.assemble_prototype(self)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


class DependencyNotFoundError(Exception):
    def __init__(self, msg=None, status_code=None):
        message = "{} {}".format(msg or "",
                                 status_code or "")
        super(DependencyNotFoundError, self).__init__(message)


class FactoryNotFound(Exception):
    """Exception when factory function is not found for column"""
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class AttemptLimitReached(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class NoEntryPointGroupError(Exception):
    """Exception when no entry_point_group has been set"""
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)

class ResourceNotFoundError(Exception):
    """Exception when no entry_point_group has been set"""
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import logging

from knockoff.orm import get_engine
from joblib import Parallel, delayed

logger = logging.getLogger(__name__)


def _to_sql(df, engine_builder, table, **kwargs):
    engine = engine_builder.build()
    with engine.connect() as conn:
        df.to_sql(table, conn, method="multi", **kwargs)


def to_sql(sink, table, df):
    logger.info("Populating table: {}".format(table))

    chunksize = sink.config.get("chunksize", 10000)
    n_jobs = sink.config.get("n_jobs", -1)
    parallelize = sink.config.get("parallelize", True)
    database = sink.config.get('database')
    kwargs = sink.config.get('kwargs', {})
    engine_builder = get_engine(database, build='builder')
    if parallelize:
        Parallel(n_jobs=n_jobs)(delayed(_to_sql)(df[i:i+chunksize],
                                                 engine_builder,
                                                 table,
                                                 **kwargs)
                                for i in range(0, len(df),
                                               chunksize))
    else:
        _to_sql(df, engine_builder, table, **kwargs)

    logger.info("Populated table: {}".format(table))


def to_parquet(sink, table, df):
    logger.info("Writing table ({}) to parquet.".format(table))
    df.to_parquet(sink.config['fname'],
                  **sink.config.get('kwargs', {}))
    logger.info("Wrote table ({}) to parquet.".format(table))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['knockoff',
 'knockoff.command',
 'knockoff.factory',
 'knockoff.sdk',
 'knockoff.sdk.container',
 'knockoff.sdk.factory',
 'knockoff.sdk.factory.next_strategy',
 'knockoff.tempdb',
 'knockoff.utilities',
 'knockoff.utilities.date',
 'knockoff.utilities.orm',
 'knockoff.writer']

package_data = \
{'': ['*']}

install_requires = \
['dependency_injector>=4.34.0,<4.35.0',
 'dotty_dict>=1.2.1',
 'enum34>=1.1.9',
 'faker>=3.0.1',
 'ipython>=5.9.0',
 'joblib>=0.14.1',
 'networkx>=2.2',
 'numpy>=1.16.6',
 'pandas>=0.24.2',
 'psycopg2>=2.8.4',
 'pyaml>=19.12.0',
 'pyarrow>=0.15.1',
 's3fs>=0.2.2',
 'six>=1.12.0',
 'sqlalchemy-utils>=0.32.12',
 'testing.postgresql>=1.3.0']

extras_require = \
{'complete': ['Pyrseas>=0.9.0']}

entry_points = \
{'console_scripts': ['knockoff = knockoff.cli_v2:main'],
 'knockoff.cli.command': ['legacy = knockoff.cli:main',
                          'run = knockoff.command.run:main',
                          'version = knockoff.command.version:main'],
 'knockoff.factory.component.function': ['numpy.random.poisson = '
                                         'numpy.random:poisson'],
 'knockoff.factory.sink.dump_strategy': ['noop = knockoff.utilities.mixin:noop',
                                         'parquet = '
                                         'knockoff.writer.pandas:to_parquet',
                                         'sql = knockoff.writer.pandas:to_sql'],
 'knockoff.factory.source.component.load_strategy': ['autoincrement = '
                                                     'knockoff.factory.component:load_autoincrement',
                                                     'faker = '
                                                     'knockoff.factory.counterfeit:load_faker_component_generator',
                                                     'function = '
                                                     'knockoff.utilities.mixin:noop',
                                                     'knockoff = '
                                                     'knockoff.utilities.mixin:noop'],
 'knockoff.factory.source.part.load_strategy': ['cartesian-product = '
                                                'knockoff.factory.part:cartesian_product_strategy',
                                                'concat = '
                                                'knockoff.factory.part:concat_strategy',
                                                'faker = '
                                                'knockoff.factory.counterfeit:load_faker',
                                                'inline = '
                                                'knockoff.factory.part:read_part_inline',
                                                'io = '
                                                'knockoff.io:load_strategy_io',
                                                'period = '
                                                'knockoff.factory.part:generate_part_periods'],
 'knockoff.factory.source.prototype.load_strategy': ['components = '
                                                     'knockoff.factory.prototype:load_prototype_from_components',
                                                     'concat = '
                                                     'knockoff.factory.part:concat_strategy',
                                                     'io = '
                                                     'knockoff.io:load_strategy_io'],
 'knockoff.factory.source.table.load_strategy': ['io = '
                                                 'knockoff.io:load_strategy_io',
                                                 'knockoff = '
                                                 'knockoff.factory.table:load_knockoff'],
 'knockoff.io.readers': ['inline = knockoff.io:read_inline',
                         'pandas.read_csv = pandas:read_csv',
                         'pandas.read_json = pandas:read_json',
                         'pandas.read_parquet = pandas:read_parquet',
                         'read_multi_parquet = knockoff.io:read_multi_parquet',
                         'sql = knockoff.io:read_sql']}

setup_kwargs = {
    'name': 'knockoff',
    'version': '4.0.0',
    'description': 'Library for generating and bootstrapping mock data',
    'long_description': 'Knockoff Factory\n---\n[![codecov](https://codecov.io/gh/Nike-Inc/knockoff-factory/branch/master/graph/badge.svg?token=93wOmtZxIk)](https://codecov.io/gh/Nike-Inc/knockoff-factory)\n[![Test](https://github.com/Nike-Inc/knockoff-factory/actions/workflows/python-test.yaml/badge.svg)](https://github.com/Nike-Inc/knockoff-factory/actions/workflows/python-test.yaml) \n[![PyPi Release](https://github.com/Nike-Inc/knockoff-factory/actions/workflows/python-build.yaml/badge.svg)](https://github.com/Nike-Inc/knockoff-factory/actions/workflows/python-build.yaml) \n[![Docker Build](https://github.com/Nike-Inc/knockoff-factory/actions/workflows/docker-build.yaml/badge.svg)](https://github.com/Nike-Inc/knockoff-factory/actions/workflows/docker-build.yaml)\n![License](https://img.shields.io/pypi/l/knockoff)\n![Python Versions](https://img.shields.io/pypi/pyversions/knockoff)\n![Docker Image Size](https://img.shields.io/docker/image-size/nikelab222/knockoff-factory/latest)\n![Python Wheel](https://img.shields.io/pypi/wheel/knockoff)\n\nA library for generating mock data and creating database fixtures that can be used for unit testing.\n\n# Run poetry install/update with psycopg2\n\nRequirements:\n* postgresql (`brew install postrgresql`)\n\nRun the following command:\n```shell script\npg_config | grep "LDFLAGS ="\n```\nOutput:\n> LDFLAGS = -L/usr/local/opt/openssl@1.1/lib -L/usr/local/opt/readline/lib -Wl,-dead_strip_dylibs\n\nTake the value of `LDFLAGS` and set that environment variable. E.g.:\n```shell script\nexport LDFLAGS="-L/usr/local/opt/openssl@1.1/lib -L/usr/local/opt/readline/lib -Wl,-dead_strip_dylibs"\n``` \nYou should now be able to run poetry install and/or update commands without failing on psycopg2.\n\n\n### Local Postgres Setup\nThe following steps can be used to setup a local postgres instance for testing.\n\n#### Requirements\n* docker\n* poetry (`curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python`)\n* postgresql (`brew install postgresql`) or pgcli (`brew install pgcli`) \n  \n#### Run Postgres\n1. Pull docker image `docker pull postgres:11.7`\n2. Run docker container: `docker run --rm  --name pg-docker -e POSTGRES_PASSWORD=docker -d -p 5432:5432  postgres:11.7` \n    * Note: you can see the running container with `docker ps` and can terminate it with `docker kill pg-docker`\n\nYou can now access the shell with `PGPASSWORD=docker pgcli -h localhost -U postgres` or\n`PGPASSWORD=docker psql -h localhost -U postgres`.\n    \n\n## Tests\nRun unit tests\n```bash\npoetry run pytest\n```\n\nThe unit tests depend on fixtures using ephemeral postgres databases\nand/or instances. By default it will attempt to connect to an existing\ninstance at `postgresql://postgres@localhost:5432/postgres` and will\ncreate and destroy databases per fixture. This postgres location can\nbe overridden with the `KNOCKOFF_TEST_DB_URI` environment variable.\n\nIf no external postgres instance is available for testing, but postgresql is\ninstalled, the `TEST_USE_EXTERNAL_DB` environment variable can be set to `0`.\nThe fixtures will then rely on the `testing.postgresql` library to create\nephemeral postgres instances per fixture.\n\nIf postgres is not available, dependent tests can be disabled with the\nfollowing environment variable\n```bash\nexport TEST_POSTGRES_ENABLED=0\n```\n\n### Knockoff Python Based Configuration\n\nWe will be providing documentation along with example repo\'s for\nthe python based configuration using the knockoff.sdk modules.\n\n\n### Knockoff Legacy Yaml Based Configuration\nNote: This yaml based configuration has been moved under the legacy subcommand\nfor knockoff and a new yaml based configuration will be introduced that \nrelies on the same objects in python based configuartion (knockoff.sdk). \n\n#### Creating Databases\nKnockoff will start by creating any specified databases. This section\nis optional if you do not need a database created. You can also configure\nan engine builder to use by providing the `config` parameter for each\nconfigured database (factory function used is knockoff.utilities.orm.sql.EngineBuilder.from_config)\notherwise the default engine based on knockoff environment variables will be used.\nThe following yaml will result in the following sql queries:\n* `create database mydb;`\n* `create user myuser with encrypted password \'MYUSER_PASSWORD\';`\n    * `MYUSER_PASSWORD` is replaced with the value of the corresponding environment variable\n* `grant all privileges on database mydb to myuser;`\n```yaml\ncreate-databases:\n  - name: mydb\n    type: postgres\n    users:\n      - user: myuser\n        password_env: MYUSER_PASSWORD\n```\n\n#### Load existing table definitions\nKnockoff uses [Pyrseas](https://github.com/perseas/Pyrseas)\'s yamltodb tool to load existing table\ndefinitions into a database.\n\n##### Example:\nIn this example there is a table `films` with the following definition:\n```commandline\n+----------+-------------------+-------------+\n| Column   | Type              | Modifiers   |\n|----------+-------------------+-------------|\n| title    | character varying |  not null   |\n| director | character varying |             |\n| year     | character varying |             |\n+----------+-------------------+-------------+\nIndexes:\n    "films_pkey" PRIMARY KEY, btree (title)\n```\nExecuting `dbtoyaml mydb` results in the following yaml that knockoff\ncan be configured to use to load with `yamltodb`. \n```yaml\nschema public:\n  description: standard public schema\n  owner: postgres\n  privileges:\n  - PUBLIC:\n    - all\n  - postgres:\n    - all\n  table films:\n    columns:\n    - title:\n        not_null: true\n        type: character varying\n    - director:\n        type: character varying\n    - year:\n        type: character varying\n    owner: myuser\n    primary_key:\n      films_pkey:\n        columns:\n        - title\n```\nNote: If you are running the local postgres setup described above and running from within a docker container on your mac, you can use the following: `PGPASSWORD=docker dbtoyaml -H docker.for.mac.host.internal -U postgres mydb`\n\n#### Loading data into tables\nData can be loaded into new or existing tables.\n\n##### Examples\nThe following example loads data into an existing table from a provided csv.\n```yaml\nknockoff:\n  dag:\n    - name: films # arbitrary name of node in dag\n      type: table # table | prototype | component | part\n      table: films # defaults to the name of the node if not provided \n      source:\n        strategy: io\n        reader: pandas.read_csv\n        kwargs:\n          filepath_or_buffer: example/films.csv # local or s3:// path\n          sep: "|"\n      sink:\n        database: mydb\n        kwargs:\n          if_exists: append # defaults to fail\n          index: false # Data is loaded into a pandas DataFrame this option ignores the index\n```\n\nThe following example loads data into a new table from data defined in the yaml.\n```yaml\nknockoff:\n  dag:\n    - name: films2 # Note: "table" key not specified, so defaults to "film2"\n      type: table\n      source:\n        strategy: io\n        reader: inline\n        kwargs:\n          sep: ","\n          data: |\n            title,director,year\n            t5,d1,2020\n            t6,d2,2020\n            t7,d1,2020\n      sink:\n        database: mydb\n        user: myuser\n        password_env: MYUSER_PASSWORD\n        kwargs:\n          index: false\n```\n\n#### Generating fake retail data\nKnockoff uses [faker](https://github.com/joke2k/faker) to help generate fake retail data that can be used for testing.\nHierarchical relationships with various dependencies can be also be modelled with knockoff. This [example](examples/knockoff.yaml)\ngenerates the following tables (in addition to the above examples).\n```shell script\npostgres@localhost:mydb> select * from location;\n+-------------------------------+---------------+-----------+\n| address                       | location_id   | channel   |\n|-------------------------------+---------------+-----------|\n| 07528 Fischer Track Suite 779 | 1             | nfs       |\n| Melissaview, MD 90363         |               |           |\n| 1535 Kelly Canyon             | 2             | nso       |\n| Rhodesborough, CA 43893       |               |           |\n| 216 Kayla Lake Apt. 126       | 3             | nso       |\n| South Matthewmouth, OH 36332  |               |           |\n| 561 Jones Burg Suite 382      | 4             | nso       |\n| Hugheschester, DE 21908       |               |           |\n| 042 Robinson Fort Suite 945   | 5             | nfs       |\n| Pattersonshire, NC 96317      |               |           |\n| 2332 Watkins Road             | 0             | digital   |\n| Davidfort, MS 71411           |               |           |\n+-------------------------------+---------------+-----------+\n\npostgres@localhost:mydb> select * from product;\n+------------+----------+-----------------------+---------------+------------+\n| division   | gender   | category              | color         | sku        |\n|------------+----------+-----------------------+---------------+------------|\n| apparel    | men      | shorts                | PaleGoldenRod | 6357812379 |\n| apparel    | women    | pants & tights        | NavajoWhite   | 8332320303 |\n| apparel    | men      | tops & t-shirts       | Lavender      | 9243289077 |\n| shoes      | men      | lifestyle             | PaleGoldenRod | 7270972977 |\n| apparel    | women    | pants & tights        | PaleGoldenRod | 4443641793 |\n| apparel    | women    | hoodies & sweatshirts | NavajoWhite   | 6130018459 |\n| shoes      | women    | jordan                | Lavender      | 3791231041 |\n| apparel    | men      | pants & tights        | PaleGoldenRod | 3899370297 |\n| apparel    | men      | shorts                | Lavender      | 7557742055 |\n| apparel    | men      | pants & tights        | SkyBlue       | 9785957221 |\n| apparel    | women    | shorts                | SkyBlue       | 9979359561 |\n| apparel    | women    | tops & t-shirts       | Lavender      | 7006056836 |\n| shoes      | women    | jordan                | Lavender      | 4853474331 |\n| shoes      | women    | jordan                | NavajoWhite   | 6589395336 |\n| apparel    | men      | pants & tights        | Beige         | 7168664719 |\n| apparel    | men      | hoodies & sweatshirts | Beige         | 7525844204 |\n| apparel    | men      | shorts                | SkyBlue       | 9735336861 |\n| shoes      | men      | skateboarding         | SkyBlue       | 6385212885 |\n| apparel    | men      | tops & t-shirts       | Beige         | 9735107927 |\n| apparel    | women    | pants & tights        | SkyBlue       | 2633853831 |\n| apparel    | women    | jackets & vests       | NavajoWhite   | 2758275877 |\n| apparel    | men      | shorts                | Lavender      | 1330756304 |\n| apparel    | women    | tops & t-shirts       | NavajoWhite   | 9334676293 |\n| shoes      | men      | skateboarding         | Lavender      | 6735393792 |\n| apparel    | men      | jackets & vests       | Lavender      | 2907811814 |\n+------------+----------+-----------------------+---------------+------------+\n\npostgres@localhost:mydb> select * from transactions limit 10;\n+---------------+------------+-----------+------------+------------+------------+\n| location_id   | sku        | line_id   | order_id   | quantity   | date       |\n|---------------+------------+-----------+------------+------------+------------|\n| 5             | 7557742055 | 2         | 2957859949 | 0          | 2018-05-06 |\n| 1             | 1330756304 | 1         | 3920316859 | 0          | 2018-07-19 |\n| 4             | 9243289077 | 3         | 1875617688 | 0          | 2019-10-14 |\n| 3             | 9334676293 | 2         | 7317451987 | 0          | 2018-06-09 |\n| 0             | 9979359561 | 3         | 1236640244 | 2          | 2019-07-17 |\n| 0             | 9735107927 | 1         | 9030486883 | 3          | 2018-04-27 |\n| 1             | 9735336861 | 2         | 6196902209 | 2          | 2020-01-21 |\n| 3             | 7006056836 | 4         | 1432537227 | 0          | 2019-04-22 |\n| 2             | 2907811814 | 1         | 4039132536 | 0          | 2019-09-23 |\n| 0             | 4443641793 | 5         | 8648324533 | 2          | 2018-09-06 |\n+---------------+------------+-----------+------------+------------+------------+\n```\n\n#### Run the example:\n\n1. Run postgres: `docker run --rm  --name pg-docker -e POSTGRES_PASSWORD=docker -d -p 5432:5432  postgres:11.9`\n    * Terminate existing container with `docker kill pg-docker`\n    * Note: The example assumes you\'re running with *POSTGRES_PASSWORD=docker* and on port *5432*\n2. Checkout the repo or download the examples folder\n3. Pull knockoff docker image: `docker pull knockoff-factory`\n```commandline\ndocker run --rm -v $PWD/examples:/examples \\\n-e KNOCKOFF_DB_HOST=\'docker.for.mac.host.internal\' \\\n-e KNOCKOFF_DB_USER=\'postgres\' \\\n-e KNOCKOFF_DB_PASSWORD=\'docker\' \\\n-e KNOCKOFF_DB_NAME=\'knockoff\' \\\n-e KNOCKOFF_CONFIG=/examples/knockoff.yaml knockoff-factory:latest knockoff legacy\n```\nNote: if you are loading data from an s3 bucket you have access to, you can enable your docker\ncontainer access to those credentials by adding `-v ~/.aws:/root/.aws` to the `docker run` command.\n\n\n### Future work\n* Add documentation and examples for SDK\n* Add yaml based configuration for SDK\n* Make extensible generic output for KnockffDB.insert (csv, parquet, etc)\n* Enable append option for KnockoffDB.insert\n* Autodiscover and populate all tables by using reflection and building dependency graph with foreign key relationships\n* Documentation / use-cases\n* Parallelize execution of dag. (e.g. https://ipython.org/ipython-doc/stable/parallel/dag_dependencies.html)\n',
    'author': 'Gregory Yu',
    'author_email': 'gregory.yu@nike.com',
    'maintainer': 'Mohamed Abdul Huq Ismail',
    'maintainer_email': 'Abdul.Ismail@nike.com',
    'url': 'https://github.com/Nike-Inc/knockoff-factory',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import numpy as np
from faker import Faker


class ColumnFactory(object):
    def __init__(self, column, callable_, depends_on=None):
        self.column = column
        self.callable = callable_
        self.depends_on = depends_on

    def __call__(self, *args, **kwargs):
        return {self.column: self.callable(*args,**kwargs)}


class ChoiceFactory(object):
    def __init__(self, choices, p=None, replace=True):
        self.choices = choices
        self.p = p
        self.replace = replace

    def __call__(self, size=None, p=None, replace=None):
        if replace is None:
            replace = self.replace
        return np.random.choice(self.choices,
                                size=size,
                                replace=replace,
                                p=p or self.p)


class FakerFactory(object):
    def __init__(self, method, faker=None, **kwargs):
        self.method = method
        self.faker = faker or Faker()
        self.kwargs = kwargs

    def __call__(self):
        return getattr(self.faker, self.method)(**self.kwargs)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import os
from pprint import pformat
import logging

from faker import Faker
import pandas as pd

from knockoff.utilities.functools import call_with_args_kwargs

logger = logging.getLogger(__name__)

KNOCKOFF_ATTEMPT_LIMIT_ENV = "KNOCKOFF_ATTEMPT_LIMIT"


def load_faker_component_generator(source, assembler, name):
    fake = Faker()

    def generator():
        while True:
            yield call_with_args_kwargs(getattr(fake,
                                        source.config['method']),
                                        source.config.get('args'),
                                        source.config.get('kwargs'))
    return generator()


def load_faker(source, assembler, name):
    fake = Faker()
    data = set()
    i = 0
    limit = int(os.environ.get(KNOCKOFF_ATTEMPT_LIMIT_ENV, 1000000))
    while len(data) < source.config['number'] and i < limit:
        item = call_with_args_kwargs(getattr(fake, source.config['method']),
                                     source.config.get('args'),
                                     source.config.get('kwargs'))
        if isinstance(item, (list, tuple)):
            tuple(item)
        if item not in data:
            data.add(item)
        i += 1
    if i >= limit:
        logger.error(pformat(source.config))
        raise Exception("Attempts to create unique set reached: {}"
                        .format(limit))

    # TODO: move to property / function that lazily checks env var that be
    #       used globally for package
    KNOCKOFF_TEST_MODE = (os.environ
                          .get("KNOCKOFF_TEST_MODE", '0')
                          .lower()) in {'1', 'true', 'y', 'yes'}
    return pd.DataFrame(sorted(data)) if KNOCKOFF_TEST_MODE else pd.DataFrame(data)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import os
import logging
from datetime import datetime

from sqlalchemy import Table

import pandas as pd

from faker import Faker
from knockoff.exceptions import FactoryNotFound, AttemptLimitReached


from .factory.column import ColumnFactory


logger = logging.getLogger(__name__)


KNOCKOFF_ATTEMPT_LIMIT_ENV = "KNOCKOFF_ATTEMPT_LIMIT"


class KnockoffTable(object):

    def __init__(self, name_or_table,
                 factories=None,
                 columns=None,
                 autoload=False,
                 ignore_constraints_on_autoload=False,
                 dtype=None,
                 constraints=None,
                 default_type_factory=None,
                 faker=None,
                 attempt_limit=None,
                 size=None,
                 database_service=None,
                 table=None,
                 ):

        if isinstance(name_or_table, Table):
            self.name = name_or_table.name
            # used to set columns and dtype
            self._table = name_or_table
        else:
            self.name = name_or_table
            self._table = None

        if table is not None:
            # overrides table from name_or_table
            # intended use-case is if you want
            # to upload based on an existing table
            # definition to a table of a different name
            self._table = table

        self.factories = factories or []
        self.columns = columns
        self._is_prepared = False
        # autoload: whether or not to reflect the schema
        self.autoload = autoload
        self.dtype = dtype or {}
        self.constraints = constraints or []
        self.faker = faker or Faker()
        self.attempt_limit = attempt_limit or int(os.getenv(KNOCKOFF_ATTEMPT_LIMIT_ENV, 1000000))
        self.default_type_factory = default_type_factory or {}
        self.size = size
        self._df = None
        self.database_service = database_service
        self.ignore_constraints_on_autoload = ignore_constraints_on_autoload

    def prepare(self, lazy=True,
                database_service=None,
                autoload=None):
        """
        Precedence of columns:
        1) Provided by columns directly from __init__
        2) Provided by columns from table from __init__
        3) Provided by columns from table from autloading table by name

        Precedence of dtype:
        1) Provided by dtype directly from __init__
        2) Provided by table from __init__
        3) Provided by table autoloaded by name
        4) Default to string for column

        TODO: should we save the original init params for reset?
              or do we only allow reset of constraints?

        :param lazy:
        :return:
        """
        if lazy and self._is_prepared:
            return # if we're lazy and columns are initialized, do nothing

        dtype = {}
        columns = None

        if autoload or self.autoload:
            database_service = database_service or self.database_service
            assert database_service is not None, "DatabaseService required for autoload"
            table = database_service.reflect_table(self.name)
            columns, types = zip(*[(c.name, c.type.python_type) for c in table.c])
            dtype.update(dict(zip(columns, types)))

            if not self.ignore_constraints_on_autoload:
                constraints = database_service.reflect_unique_constraints(self.name)
                self.constraints.extend(constraints)

        if self._table is not None:
            columns, types = zip(*[(c.name, c.type.python_type) for c in self._table.c])
            dtype.update(dict(zip(columns, types)))

        self._columns = self._columns or columns
        assert self._columns is not None

        # dtype provided by init take precedence
        dtype.update(self.dtype)
        self.dtype = dtype

        self._is_prepared = True

    @property
    def columns(self):
        # TODO should we instead assert that table has been prepared?
        self.prepare(lazy=True)
        return self._columns

    @columns.setter
    def columns(self, columns):
        self._columns = columns

    @property
    def df(self):
        if self._df is None:
            self.build() # TODO: do we want to do this?
        return self._df

    @property
    def default_type_factory(self):
        return self._default_type_factory

    @default_type_factory.setter
    def default_type_factory(self, default_type_factory):
        # defaults
        self._default_type_factory = {
            int: self.faker.pyint,
            float: self.faker.pyfloat,
            str: self.faker.pystr_format,
            bool: self.faker.pybool,
            datetime: self.faker.date_time,
            dict: lambda: {}
        }
        self._default_type_factory.update(default_type_factory)

    def copy_factory(self, columns=None, rename=None):
        columns = columns or self.columns
        rename = rename or {}

        def factory():
            record = self.build_record()
            out = []
            for col in columns:
                out[rename.get(col, col)] = record[col]
            return out
        return factory

    def build_record(self):
        """TODO: should this be public?"""
        data = {}
        for factory in self.factories:
            key_values = {}
            if isinstance(factory, (tuple, list)):
                logger.warning("Using a tuple will be "
                               "deprecated in newer releases."
                               "Use ColumnFactory instead")
                col, factory = factory
                key_values = {col: factory()}
            elif (isinstance(factory, (ColumnFactory,)) and
                   factory.depends_on):
                kwargs = {key: data[key] for key in factory.depends_on}
                key_values = factory(**kwargs)
            else:
                key_values = factory()
            data.update(key_values)
        record = {}
        for col in self.columns:
            if col not in data:
                try:
                    type_ = self.dtype.get(col, str)
                    factory = self.default_type_factory[type_]
                    record[col] = factory()
                except KeyError:
                    raise FactoryNotFound("[table={}] No column factory provided"
                                          " for column={} or type={}"
                                          .format(self.name, col, type_))
            else:
                record[col] = data[col]
        return record

    def _check_constraints(self, record):
        constraints_satisfied = True
        for constraint in self.constraints:
            if not constraint.check(record):
                constraints_satisfied = False
                break
        return constraints_satisfied

    def _add_record(self, record):
        """add record to constraints if they need to track for future violations"""
        for constraint in self.constraints:
            constraint.add(record)

    def _next(self):
        attempt = 0
        while attempt < self.attempt_limit:
            attempt += 1

            record = self.build_record()

            if self._check_constraints(record):
                self._add_record(record)
                return record

        if attempt >= self.attempt_limit:
            raise AttemptLimitReached("Attempts to create df with unique "
                                      "constraint reached limit={} for table={}"
                                      .format(self.attempt_limit, self.name))

    def build(self, size=None):
        size = size or self.size
        assert size is not None
        # TODO: do this more memory efficiently?
        self._df = pd.DataFrame([self._next() for _ in range(size)])
        return self.df

    def reset(self):
        self._df = None
        for constraint in self.constraints:
            constraint.reset()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import logging

from knockoff.utilities.mixin import ResourceLocatorMixin


logger = logging.getLogger(__name__)


class TableSourceFactory(ResourceLocatorMixin, object):
    entry_point_group = "knockoff.factory.source.table.load_strategy"


class PartSourceFactory(ResourceLocatorMixin, object):
    entry_point_group = "knockoff.factory.source.part.load_strategy"


class ComponentSourceFactory(ResourceLocatorMixin, object):
    entry_point_group = "knockoff.factory.source.component.load_strategy"


class PrototypeSourceFactory(ResourceLocatorMixin, object):
    entry_point_group = "knockoff.factory.source.prototype.load_strategy"


TYPE_TO_FACTORY = {
    "part": PartSourceFactory,
    "table": TableSourceFactory,
    "component": ComponentSourceFactory,
    "prototype": PrototypeSourceFactory,
}


class KnockoffSource:
    def __init__(self, node_type, config):
        self.node_type = node_type
        self.config = config
        self.load_strategy = (TYPE_TO_FACTORY[self.node_type]()
                              .get_resource(self.config['strategy']))

    def load(self, assembler, node_name):
        return self.load_strategy(self, assembler, node_name)


class SourceMixin:
    source = None
    data = None
    name = None

    def load(self, assembler):
        if self.name is None:
            raise Exception("Missing name")
        if self.source is None:
            raise Exception('Missing source')
        if self.data is None:
            self.data = self.source.load(assembler,
                                         self.name)



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.

import logging
import re
import networkx as nx
from networkx.algorithms import topological_sort
from networkx.algorithms import is_directed_acyclic_graph


from knockoff.factory.sink import KnockoffSink
from knockoff.factory.source import KnockoffSource
from knockoff.factory.node import Table, FactoryPart, FactoryComponent, FactoryPrototype
from knockoff.exceptions import DependencyNotFoundError

logger = logging.getLogger(__name__)

DEPENDENCY_REGEX = (r'^(part|component|prototype)\:'
                    '([A-Za-z0-9-_]+)[.]?([A-Za-z0-9-_]*)$')

DEPENDENCY_PATTERN = re.compile(DEPENDENCY_REGEX)


class BlueprintBuilder:
    def __init__(self):
        self.dag = nx.DiGraph()
        self.nodes = []
        self.index = {}
        self.dependency_queue = []

    def _add_node_to_index(self, node_type, name):
        i = len(self.index)
        if (node_type, name) in self.index:
            raise Exception("Node of type={} with name={} already exists."
                            .format(node_type, name))
        self.index[(node_type, name)] = i
        return i

    def add_part(self, node):
        node_type = node['type']
        name = node['name']
        source = KnockoffSource(node_type, node['source'])

        i = self._add_node_to_index(node_type, name)

        node = FactoryPart(name, i, source)

        self.nodes.append(node)
        self.dag.add_node(i)

        dependencies = source.config.get('dependencies') or []
        for dep in dependencies:
            other_node_type, other_name, _ = Assembler.parse_dependency(dep)
            self._add_dependency(i, other_node_type, other_name)

    def _add_dependency(self, i, other_type, other_name, queue_on_error=True):
        try:
            other_i = self.index[other_type, other_name]
            self.dag.add_edge(other_i, i)
        except KeyError:
            if queue_on_error:
                self.dependency_queue.append((i, (other_type, other_name)))
            else:
                raise DependencyNotFoundError("Dependency not found: "
                                              "node={}, name={}"
                                              .format(other_type,
                                                      other_name))

    def add_table(self, node):
        node_type = node['type']
        name = node['name']
        source = KnockoffSource(node_type, node['source'])
        sink = KnockoffSink(node.get('sink', {'strategy': 'noop'}))

        i = self._add_node_to_index(node_type, name)

        node = Table(name, i, source, sink, table=node.get('table'))

        self.nodes.append(node)
        self.dag.add_node(i)

        if source.config['strategy'] == "knockoff":
            other_name = source.config.get('kwargs', {}).get('prototype')
            self._add_dependency(i, 'prototype', other_name)

    def add_prototype(self, node):
        node_type = node['type']
        name = node['name']

        i = self._add_node_to_index(node_type, name)
        source = KnockoffSource(node_type, node['source'])
        prototype = FactoryPrototype(name, i, source, unique=node.get('unique'))
        self.nodes.append(prototype)
        self.dag.add_node(i)

        if source.config['strategy'] == 'components':
            components = []
            for component in source.config['components']:
                component_index_name = "{}.{}".format(name,
                                                      component['name'])
                other_i = self._add_node_to_index('component',
                                                  component_index_name)
                component_source = KnockoffSource("component",
                                                  component['source'])
                component = FactoryComponent(component['name'],
                                             other_i,
                                             component_source)
                components.append(component)
                self.nodes.append(component)
                self._add_dependency(i, 'component',
                                     component_index_name)
                for dep in component.source.config.get('dependencies', []):
                    comp_dep_type, comp_dep_name, _ = Assembler.parse_dependency(dep)
                    self._add_dependency(other_i, comp_dep_type, comp_dep_name)

            prototype.source.components = components
        elif source.config['strategy'] == 'concat':
            for dep in source.config['dependencies']:
                other_type, other_name, _ = Assembler.parse_dependency(dep)
                self._add_dependency(i, other_type, other_name)
        elif source.config['strategy'] == 'io':
            # TODO: anything to do?
            pass
        else:
            raise ValueError("Unrecognized strategy: {} for prototype".format(source.config['strategy']))

    def build(self):
        assert len(self.index) == len(self.nodes), 'Mismatch in index and nodes'
        assert len(self.dag.nodes) == len(self.nodes), 'Mismatch in dag and nodes'
        for i, (node_type, name) in self.dependency_queue:
            self._add_dependency(i, node_type, name, queue_on_error=False)
        assert is_directed_acyclic_graph(self.dag)
        return Blueprint(self.dag, self.nodes, self.index)


class Blueprint:
    def __init__(self, dag, nodes, index):
        self.dag = dag
        self.nodes = nodes
        self.index = index

    def get_node(self, node_type, name):
        return self.nodes[self.index[node_type, name]]

    def iter_topologically(self):
        return topological_sort(self.dag)

    @staticmethod
    def from_config(config):
        # TODO: add validation for config

        builder = BlueprintBuilder()

        for node in config['dag']:

            if node['type'] == FactoryPart.node_type:
                builder.add_part(node)
            elif node['type'] == Table.node_type:
                builder.add_table(node)
            elif node['type'] == FactoryPrototype.node_type:
                builder.add_prototype(node)
            elif node['type'] == FactoryComponent.node_type:
                # TODO: can we get here? should components be restricted to prototypes
                raise Exception('component must be part of prototype')
            else:
                raise Exception('node type not recognized: {}'
                                .format(node['type']))
        blueprint = builder.build()

        return blueprint


class Assembler:
    def __init__(self, blueprint):
        self.blueprint = blueprint

    def accept(self, visitor):
        visitor.visit(self)

    def start(self):
        # TODO: we can parallelize the execution of the dag
        for i in self.blueprint.iter_topologically():
            node = self.blueprint.nodes[i]
            logger.debug(("Assembler accepting visit from node:"
                          " name={}, type={}, ix={}")
                         .format(node.name,
                                 # TODO: replace with enum / classmap?
                                 node.source.node_type,
                                 node.ix))
            self.accept(node)

    def assemble_table(self, node):
        node.load(self)
        node.sink.dump(node.table, node.data)

    def assemble_part(self, node):
        node.load(self)

    def prepare_component(self, node):
        node.prepare(self)

    def assemble_prototype(self, node):
        node.load(self)

    @staticmethod
    def parse_dependency(string, pattern=DEPENDENCY_PATTERN):
        match = pattern.match(string)
        if match:
            return match.groups()
        raise Exception("Dependency pattern not recognized: {}"
                        .format(string))






# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


import os
import argparse
import logging
import sys

from .utilities.mixin import ResourceLocatorMixin


logger = logging.getLogger(__name__)


class KnockoffCLI(ResourceLocatorMixin):
    entry_point_group = "knockoff.cli.command"

    @property
    def usage(self):
        _usage = ["knockoff <command> [<args>]",
                  "The most commonly used knockoff commands are:"]

        description = {
            "legacy": ("Load data into a database "
                       "based on the legacy yaml "
                       "configuration"),
            "run": "Load data into a database given "
                   "a knockoff_db path",
            "version": "Print knockoff version",
        }

        for key in sorted(self.get_resource_names()):
            try:
                _usage.append("{command}\t\t{description}".format(
                    command=key, description=description[key]
                ))
            except KeyError:
                raise Exception(f"Missing usage description for {key}")

        return "\n".join(_usage)

    def __init__(self):
        parser = argparse.ArgumentParser(
            description="knockoff cli",
            usage=self.usage
        )
        parser.add_argument('command',
                            choices=self.get_resource_names(),
                            help="Subcommand to run",
                            metavar="COMMAND")
        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2])
        command = self.get_resource(args.command)
        command(sys.argv[2:])


def setup_logger(verbose=False):
    level = logging.DEBUG if verbose else (os.environ
                                             .get('KNOCKOFF_LOG_LEVEL',
                                                  'INFO').upper())
    format_ = os.environ.get('KNOCKOFF_LOG_FORMAT',
                            ('[%(asctime)s] [%(name)s] [%(processName)s]'
                             ' [%(levelname)s]: %(message)s'))
    logging.basicConfig(format=format_,
                        level=level)


def main():
    setup_logger()
    KnockoffCLI()


if __name__ == "__main__":
    main()



# Copyright 2021-present, Nike, Inc.
# All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in
# the LICENSE file in the root directory of this source tree.


from .df import sample_df, cycle_df_factory


def sample_table(table, **kwargs):
    return sample_df(table.df, **kwargs)


def cycle_table_factory(table):
    return cycle_df_factory(table.df)
